/**

   Our main file

*/
// We need the Adafruit Arcada Library for our basic functionality
#include <Adafruit_Arcada.h>
#include "words.h"
#include "images.h"
#include "penguin.h"
#include "monsters.h"

// Global variables
Adafruit_Arcada arcada;
#define START_SCREEN 0
#define COLLECTING 1
#define GUESSING 2
#define GUESS_CORRECT 3
#define GUESS_WRONG 4
#define PLAYER_HIT 5
int state = START_SCREEN;

//Globals for music
File musicFile;
bool musicIsPlaying = false;
int MusikCounter = 1;
bool changeMusic = true;

//Music Methode 
void loadMusic(int musicCounter){
/*
  if (musicCounter == 1){
    musicFile = arcada.openFileByIndex("/musicStart", 0, O_READ, "wav");
  }
  else if (musicCounter ==2){
    musicFile = arcada.openFileByIndex("/musicCollecting", 0, O_READ, "wav");
  }
  else if (musicCounter ==3){
    musicFile = arcada.openFileByIndex("/musicHit", 0, O_READ, "wav");
  }
  else if (musicCounter ==4){
    musicFile = arcada.openFileByIndex("/musicLose", 0, O_READ, "wav");
  }
  else if (musicCounter ==5){
    musicFile = arcada.openFileByIndex("/musicWin", 0, O_READ, "wav");
  }
  else {
    musicFile = arcada.openFileByIndex("/music", 0, O_READ, "wav");
  }*/
  musicFile = arcada.openFileByIndex("/music", 0, O_READ, "wav");
  if (!musicFile){
    errorReporter("Fatal error: No sound file present");
    }
    uint32_t sampleRate;
    
    //load wav information, enable spakers and set interrupt for samplings
    wavStatus status = arcada.WavLoad(musicFile, &sampleRate);
    if((status == WAV_LOAD) || (status == WAV_EOF)){
      arcada.enableSpeaker(true);
      arcada.timerCallback(sampleRate, timedInterruptCallback);
      }else {
        errorReporter("Fatal error: No sound file present");
        }
}
//Declaration of InterruptCallback (for Music)
void timedInterruptCallback(void) {
  // we just entered the playing state
  if (!musicIsPlaying) {
  }
  // play next sample
  wavStatus status = arcada.WavPlayNextSample();

  // update timing information
  // update playing state and hardware depending on wav playing status
  if (status == WAV_EOF) {
    arcada.timerStop();
    arcada.enableSpeaker(false);
    musicIsPlaying = false;
    musicFile.close();
  } else {
    musicIsPlaying = true;
  }
}
// Change this for a different word
const int numberOfLetters = 4;
char letters[] = "ALEF"; //{ 'A', 'L', 'E', 'F', '\0' };
char correctWord[] = "LEAF";
const int wordLength = 4;

// For guessing words
//bool gameStopped = false; // Is the guess currently being checked?
bool guessedCorrectly = true; // Has the last guess been correct?
int guessesLeft = 3; // How many wrong guesses are still allowed?

// For collecting letters & Player 
int xposition=10; // x-position of player
int yposition=10; // y-position of player
#define MOVING_CYCLE_LENGTH 4;
int stepincycle=0;
#define MOVING_DOWN 0
#define MOVING_UP 1
#define MOVING_LEFT 2
#define MOVING_RIGHT 3
int movingdirection=MOVING_LEFT;

bool lettersCollected[] = {false, false, false, false};
int numberOfLettersRemaining = 4;
int xposOfLetters[] = {30, 60, 80, 110};
int yposOfLetters[] = {80, 20, 100, 40};

int numberOfEnemies = 4;
int xposOfEnemies[] = {30, 60, 80, 110};
int yposOfEnemies[] = {80, 20, 100, 40};
bool enemyOnWayBack[] = {false, false, false, false};

// Global variables for user interaction
int choices[wordLength];
int letterBeingChosen = 0;

void initializeWord() { 
  const char* wordFromFile = words[random() % wordsWeCanChooseFrom]; // determine random word

  // Now convert wordFromFile to upper case and shuffle it
  int j = 0;
  char ch;
  int swapwith = 0;

  while (wordFromFile[j]) {
    ch = wordFromFile[j];
    correctWord[j] = toupper(ch);
    if (j == 0) {
      swapwith = 0;
    } else {
      swapwith = random() % j;
    }
    letters[j] = letters[swapwith];
    letters[swapwith] = correctWord[j];
    j++;
  }
}

// Level data easily accessible to be able to change everything fast 
// and every code after this is still right
//globale Variablen der Map
const uint16_t levelWidth = 9;
const uint16_t levelHeight = 7;
const uint16_t levelTileSize = 20;
GFXcanvas16* levelCanvas = new GFXcanvas16(levelWidth * levelTileSize, levelHeight * levelTileSize);
//Map
 const uint8_t levelData[levelWidth * levelHeight] = {
  
  3, 3, 3, 3, 3, 3, 3, 3, 3,
  3, 3, 3, 3, 3, 3, 3, 3, 3, 
  3, 3, 2, 2, 2, 3, 3, 3, 3, 
  3, 3, 2, 2, 2, 3, 3, 3, 3, 
  3, 3, 3, 3, 3, 3, 3, 2, 3, 
  3, 3, 3, 3, 3, 3, 3, 3, 3,
  3, 3, 3, 3, 3, 3, 3, 3, 3, 
  
};

// Here tileData is just color values - to draw images just save a
// pointer to the bitmap here and change the drawing function later
const uint16_t levelTileData[4] = {
  ARCADA_RED,
  ARCADA_GREEN,
  ARCADA_BLUE,
  ARCADA_WHITE,
};

bool isCollision(uint8_t new_x, uint8_t new_y){ 
  uint8_t nw = ((new_y + levelTileSize / 2) / levelTileSize) * levelWidth + ((new_x - levelTileSize / 2) / levelTileSize);
  uint8_t ne = ((new_y + levelTileSize / 2) / levelTileSize) * levelWidth + ((new_x + levelTileSize / 2) / levelTileSize);
  uint8_t sw = ((new_y - levelTileSize / 2) / levelTileSize) * levelWidth + ((new_x - levelTileSize / 2) / levelTileSize);
  uint8_t se = ((new_y - levelTileSize / 2) / levelTileSize) * levelWidth + ((new_x + levelTileSize / 2) / levelTileSize);

  if (levelData[nw] != 2 && levelData[ne] != 2 && levelData[sw] != 2 && levelData[se] != 2) {
    return false;
  } else {
    return true;
  }
}

void initializeMap(){
 
  for(int y = 0; y < levelHeight; y++) {
    for(int x = 0; x < levelWidth; x++) {
      uint16_t levelDataIndex = y * levelWidth + x;
      uint16_t levelTileDataIndex = levelData[levelDataIndex];
      uint16_t tileColor = levelTileData[levelTileDataIndex];
      levelCanvas->fillRect(x * levelTileSize, y * levelTileSize, levelTileSize, levelTileSize, tileColor);
    }
  }
}

void initializeGame() { 
  initializeWord();
  for(int i=0; i<numberOfLetters; i++) {
    do
    {
      lettersCollected[i] = false;
      xposOfLetters[i] = random() % 130; 
      yposOfLetters[i] = random() % 100;
    } while (isCollision(xposOfLetters[i], yposOfLetters[i]));
  }
  for(int i=0; i<numberOfEnemies; i++) {
    do
    {
      xposOfEnemies[i] = random() % 130;
      yposOfEnemies[i] = random() % 100;
    } while (isCollision(xposOfEnemies[i], yposOfEnemies[i]));
  }
  numberOfLettersRemaining = numberOfLetters;
  guessedCorrectly = false;
  guessesLeft = 3;
  letterBeingChosen = 0;
}

void rightKey() {
  if (letterBeingChosen < wordLength-1) {
    letterBeingChosen++;
  }
}

void leftKey() {
  if (letterBeingChosen > 0) {
    letterBeingChosen--;
  }
}

void downKey() {
  choices[letterBeingChosen] = ((choices[letterBeingChosen] + 1) % numberOfLetters);
}

void upKey() {
  if (choices[letterBeingChosen] == 0) {
    choices[letterBeingChosen] = numberOfLetters - 1;
  } else {
    choices[letterBeingChosen] = choices[letterBeingChosen] - 1;
  }
}

void setup() {
  // initialize arcada library
  if (!arcada.arcadaBegin())
  {
   // errorReporter("Fatal error: Arcada library failed.");
  }

  // initialize the file system on the flash (first flashing should be done via CircuitPython)
  if (!arcada.filesysBegin()) { // arcada.filesysBeginMSD() for "USB-Stick" mode
    //errorReporter("Fatal error: Filesystem failed.");
  }

  // initialize serial connection
  Serial.begin(9600);
  Serial.println("Booting into game...");

  // configure board display and enable frame buffer mode
  arcada.displayBegin();
  arcada.setBacklight(255);
  arcada.createFrameBuffer(ARCADA_TFT_WIDTH, ARCADA_TFT_HEIGHT);
  initializeMap();
  Serial.println("Initializing display...");

  Serial.println("Finished booting!");
}

void handleStartScreen() {
  
  // handle buttons
  uint8_t buttonState = arcada.readButtons();
  uint8_t justPressed = arcada.justPressedButtons();

  if (justPressed & ARCADA_BUTTONMASK_START) {
    srandom(millis()); // random seed
    initializeGame();
    changeMusic = true;
    state = COLLECTING;
  } else {
    GFXcanvas16 *canvas = arcada.getCanvas();
    canvas->fillScreen(ARCADA_WHITE);

    canvas->setTextSize(2);
    canvas->setTextColor(ARCADA_PURPLE);
    canvas->setCursor(18, 12);
    canvas->print("WORDS");
    canvas->setTextColor(ARCADA_RED);
    canvas->setCursor(28, 32);
    canvas->print("MAKE");
    canvas->setTextColor(ARCADA_BLUE);
    canvas->setCursor(38, 52);
    canvas->print("HAPPY");
    
    canvas->setTextSize(1);
    canvas->setTextColor(ARCADA_PURPLE);
    canvas->setCursor(18, 108);
    canvas->print("Press START to begin.");
  }
}

void handleGuessing() {
  GFXcanvas16 *canvas = arcada.getCanvas();
  
  // handle buttons
  uint8_t buttonState = arcada.readButtons();
  uint8_t justPressed = arcada.justPressedButtons();
 
  if (justPressed & ARCADA_BUTTONMASK_A) {
      guessedCorrectly = true;
      for (int i = 0; i < wordLength && guessedCorrectly; i++) {
        if (correctWord[i] != letters[choices[i]]) {
          guessedCorrectly = false;
        }
      }
      if (!guessedCorrectly) {
        guessesLeft--;
      }
      if(guessedCorrectly) {
        state = GUESS_CORRECT;
      } else {
        state = GUESS_WRONG;
      }
  } else {
    if (justPressed & ARCADA_BUTTONMASK_RIGHT) {
      rightKey();
    }
    if (justPressed & ARCADA_BUTTONMASK_LEFT) {
      leftKey();
    }
    if (justPressed & ARCADA_BUTTONMASK_DOWN) {
      downKey();
    }
    if (justPressed & ARCADA_BUTTONMASK_UP) {
      upKey();
    }
    
    canvas->fillScreen(ARCADA_WHITE);

    canvas->setTextSize(1);
    canvas->setTextColor(ARCADA_PURPLE);
    canvas->setCursor(18, 2);
    canvas->print("Now, build a word!");
    canvas->setTextColor(ARCADA_RED);
    canvas->setCursor(18, 34);
    canvas->print("Your letters:");
    canvas->setCursor(18, 50);
    canvas->print(letters);

    canvas->setTextColor(ARCADA_PURPLE);
    canvas->setCursor(18, 108);
    canvas->print("Press A to continue.");

    canvas->setTextSize(2);
    for (int i = 0; i < wordLength; i++) {
      canvas->setCursor(18 + i * 16, 82);
      if (i == letterBeingChosen) {
        canvas->setTextColor(ARCADA_BLUE);
      } else {
        canvas->setTextColor(ARCADA_ORANGE);
      }
      canvas->print(letters[choices[i]]);
    }
  }
}

void handleGuessCorrect() {
  
    uint8_t buttonState = arcada.readButtons();
    uint8_t justPressed = arcada.justPressedButtons();
    if (justPressed & ARCADA_BUTTONMASK_A) {
      initializeGame();
      changeMusic = true;
      state = COLLECTING;
    } else {
      GFXcanvas16 *canvas = arcada.getCanvas();
      canvas->fillScreen(ARCADA_PURPLE);
      canvas->setTextColor(ARCADA_WHITE);
      canvas->setCursor(18, 58);
      canvas->setTextSize(2);
      canvas->print("Correct!");
    }
}

void handleGuessWrong() {
  
    uint8_t buttonState = arcada.readButtons();
    uint8_t justPressed = arcada.justPressedButtons();
    if (justPressed & ARCADA_BUTTONMASK_A) {
      if(guessesLeft == 0) {
        initializeGame();
        changeMusic = true;
        state = COLLECTING;
      } else {
        changeMusic = true;
        state = GUESSING;
      }
    } else {
      GFXcanvas16 *canvas = arcada.getCanvas();
      canvas->fillScreen(ARCADA_RED);
      canvas->setTextColor(ARCADA_WHITE);
      canvas->setCursor(18, 28);
      canvas->setTextSize(2);
      canvas->print("Incorrect!");
      canvas->setCursor(18, 58);
      canvas->setTextSize(1);
      if(guessesLeft > 0) {
        char str[22];
        if(guessesLeft==1) {
          sprintf(str, "Just one more guess.", guessesLeft);
        } else {
          sprintf(str, "Just %d more guesses.", guessesLeft);
        }
        canvas->print(str);
      } else {
        char str[22];
        sprintf(str, "Your word was: %s.", correctWord);
        canvas->print(str);
        canvas->setCursor(18, 78);
        canvas->print("Let's try another word.");
      }
    }
}

void handlePlayerHit() {
  
    uint8_t buttonState = arcada.readButtons();
    uint8_t justPressed = arcada.justPressedButtons();
    if (justPressed & ARCADA_BUTTONMASK_A) {
      initializeGame();
      changeMusic = true;
      state = COLLECTING;
    } else {
      GFXcanvas16 *canvas = arcada.getCanvas();
      canvas->fillScreen(ARCADA_RED);
      canvas->setTextColor(ARCADA_WHITE);
      canvas->setCursor(18, 28);
      canvas->setTextSize(2);
      canvas->print("Oh no!");
      canvas->setCursor(18, 58);
      canvas->setTextSize(1);
      canvas->print("Start over!");
    }
}

void handleCollecting() {
 
  // player movement:
  uint8_t buttonState = arcada.readButtons();
  uint8_t new_x = xposition;
  uint8_t new_y = yposition;

  if (buttonState & ARCADA_BUTTONMASK_RIGHT) {
    movingdirection = MOVING_RIGHT;
    stepincycle = (stepincycle+1) % MOVING_CYCLE_LENGTH;
    if (xposition < ARCADA_TFT_WIDTH - 20) {
      new_x++;
    }
  }
  if (buttonState & ARCADA_BUTTONMASK_LEFT) {
    movingdirection = MOVING_LEFT;
    stepincycle = (stepincycle+1) % MOVING_CYCLE_LENGTH;
    if (xposition>0) {
      new_x--;
    }
  }
  if (buttonState & ARCADA_BUTTONMASK_DOWN) {
    movingdirection = MOVING_DOWN;
    stepincycle = (stepincycle+1) % MOVING_CYCLE_LENGTH;
    if (yposition < ARCADA_TFT_HEIGHT - 20) {
      new_y++;
    }
  }
  if (buttonState & ARCADA_BUTTONMASK_UP) {
    movingdirection = MOVING_UP;
    stepincycle = (stepincycle+1) % MOVING_CYCLE_LENGTH;
    if (yposition > 0) {
      new_y--;
    }
  }

  if (!isCollision(new_x, new_y)) {
    xposition = new_x;
    yposition = new_y;
  }

  // enemy movement
for(int i=0; i<numberOfEnemies; i++) {
    if(enemyOnWayBack[i]) {
      yposOfEnemies[i]--;
      if(yposOfEnemies[i] <= 0 || isCollision(xposOfEnemies[i], yposOfEnemies[i] - 1)) {
        enemyOnWayBack[i] = false;
      }
    } else {
      yposOfEnemies[i]++;
      if(yposOfEnemies[i] >= ARCADA_TFT_HEIGHT-12 || isCollision(xposOfEnemies[i], yposOfEnemies[i] + 1)) {
        enemyOnWayBack[i] = true;
      }
    }
    if (xposition >= xposOfEnemies[i] - 5 && xposition <= xposOfEnemies[i] + 5 && yposition >= yposOfEnemies[i] - 5 && yposition <= yposOfEnemies[i] + 5) {
      state = PLAYER_HIT;
    }
  }

  for(int i=0; i<numberOfLetters; i++) {
    if (xposition >= xposOfLetters[i] - 5 && xposition <= xposOfLetters[i] + 5 && yposition >= yposOfLetters[i] - 5 && yposition <= yposOfLetters[i] + 5) {
      if(!lettersCollected[i]) {
        numberOfLettersRemaining--;
        lettersCollected[i] = true;
      }
    }
  }

  if(numberOfLettersRemaining==0) {
    state = GUESSING;
  }
  
  // use standard rendering functions without directly updating the screen
  GFXcanvas16 *canvas = arcada.getCanvas();
  canvas->fillScreen(ARCADA_WHITE);
  canvas->drawRGBBitmap(0, 0, levelCanvas->getBuffer(), levelWidth * levelTileSize, levelHeight * levelTileSize);


  //canvas->setTextSize(2);
  //canvas->setTextColor(ARCADA_BLUE);
  //canvas->setCursor(xposition, yposition);
  //canvas->print("*");
  //canvas->drawXBitmap(xposition, yposition, pingubin_bits, pingubin_width, pingubin_height, ARCADA_BLUE);
  //canvas->drawRGBBitmap(xposition, yposition, pingubin, pingubin_mask, pingubin_width, pingubin_height);
  //canvas->drawRGBBitmap(xposition, yposition, penguin16, penguin16mask, 16, 25);
  //canvas->writeFillRect(xposition, yposition, 16, 16, ARCADA_BLUE);
  switch(movingdirection) {
    case MOVING_UP:
      switch(stepincycle) {
        case 0:
          canvas->drawRGBBitmap(xposition, yposition, penguinup2, 15, 25);
          break;
        case 1:
          canvas->drawRGBBitmap(xposition, yposition, penguinup1, 15, 25);
          break;
        case 2:
          canvas->drawRGBBitmap(xposition, yposition, penguinup2, 15, 25);
          break;
        case 3:
          canvas->drawRGBBitmap(xposition, yposition, penguinup3, 15, 25);
          break;
      }
      break;
    case MOVING_DOWN:
            switch(stepincycle) {
        case 0:
          canvas->drawRGBBitmap(xposition, yposition, penguindown2, 15, 25);
          break;
        case 1:
          canvas->drawRGBBitmap(xposition, yposition, penguindown1, 15, 25);
          break;
        case 2:
          canvas->drawRGBBitmap(xposition, yposition, penguindown2, 15, 25);
          break;
        case 3:
          canvas->drawRGBBitmap(xposition, yposition, penguindown3, 15, 25);
          break;
      }
      break;
    case MOVING_LEFT:
      switch(stepincycle) {
        case 0:
          canvas->drawRGBBitmap(xposition, yposition, penguinleft2, 15, 25);
          break;
        case 1:
          canvas->drawRGBBitmap(xposition, yposition, penguinleft1, 15, 25);
          break;
        case 2:
          canvas->drawRGBBitmap(xposition, yposition, penguinleft2, 15, 25);
          break;
        case 3:
          canvas->drawRGBBitmap(xposition, yposition, penguinleft3, 15, 25);
          break;
      }
      break;
    case MOVING_RIGHT:
      switch(stepincycle) {
        case 0:
          canvas->drawRGBBitmap(xposition, yposition, penguinright2, 15, 25);
          break;
        case 1:
          canvas->drawRGBBitmap(xposition, yposition, penguinright1, 15, 25);
          break;
        case 2:
          canvas->drawRGBBitmap(xposition, yposition, penguinright2, 15, 25);
          break;
        case 3:
          canvas->drawRGBBitmap(xposition, yposition, penguinright3, 15, 25);
          break;
      }
      break;
  }

  // draw enemies
  
  canvas->setTextSize(2);
  canvas->setTextColor(ARCADA_BLACK);
  
  for(int i=0; i<numberOfEnemies; i++) {
    switch(i){
      case 0:
        canvas->drawRGBBitmap(xposOfEnemies[i], yposOfEnemies[i], monster0, 16, 16);
        break;
      case 1:
        canvas->drawRGBBitmap(xposOfEnemies[i], yposOfEnemies[i], monster1, 16, 16);
        break;
      case 2:
        canvas->drawRGBBitmap(xposOfEnemies[i], yposOfEnemies[i], monster2, 16, 16);
        break;
      default:
        canvas->drawRGBBitmap(xposOfEnemies[i], yposOfEnemies[i], monster3, 16, 16);
        break;      
    }
    //canvas->drawRGBBitmap(xposOfEnemies[i], yposOfEnemies[i], monster0, 16, 16);
  }

  canvas->setTextSize(2);
  canvas->setTextColor(ARCADA_ORANGE);

  for(int i=0; i<numberOfLetters; i++) {
    if(!lettersCollected[i]) {
      canvas->setCursor(xposOfLetters[i], yposOfLetters[i]);
      canvas->print(letters[i]);
    }    
  }

}

void loop() {  
  switch(state) {
    case START_SCREEN:
      handleStartScreen();
      MusikCounter = 1;
      break;
    case COLLECTING:
      handleCollecting();
      MusikCounter = 2;
      break;
    case GUESSING:
      handleGuessing();
      MusikCounter = 6;
      break;
    case GUESS_CORRECT:
      handleGuessCorrect();
      MusikCounter = 5;
      break;
    case GUESS_WRONG:
      handleGuessWrong();
      MusikCounter =4;
      break;
    case PLAYER_HIT:
      handlePlayerHit();
      MusikCounter = 3;
      break;
  } 
  //music main loop
  if (musicIsPlaying) {
    // fill music buffer if needed
    if (arcada.WavReadyForData()) {
      wavStatus status = arcada.WavReadFile();
    }
  }
  if (!musicIsPlaying || changeMusic == true){
    loadMusic(MusikCounter);
    changeMusic = false;
  }
  // render the frame buffer to the screen
  arcada.blitFrameBuffer(0, 0, true, false);

  // perform standard board functions (cooperative multitasking)
  yield();
}

/*
   Minimal error reporter (adapted from Adafruit examples - without the LED blinking stuff)
*/
//Declaration of errorReporter 
void errorReporter(const char *error) {
  Serial.begin(9600);
  Serial.println(error);

  // busy wait to keep USB connections, etc. running
  while (true) {
    yield();
  }
}
